'__________________________________________________________________________________________________________________________
' Original Morse code routines based on original PICAXE CW beacon program by Steve Paige, VK6HV, (c)2008
' QRSS & FSK CW modifications/additions by Robert C. Mazur, VA3ROM (July/September 2016 & March 2017 & October 2020)
'__________________________________________________________________________________________________________________________

'__________________________________________________MORSE CODE CHARACTERS___________________________________________________
'       A .-    B -...  C -.-.  D -..    E .      F ..-.
'       G --.   H ....  I ..    J .---   K -.-    L .-..
'       M --    N -.    O ---   P .--.   Q --.-   R .-.
'       S ...   T -     U ..-   V ...-   W .--    X -..-
'       Y -.--  Z --.. 

'       1 .---- 2 ..--- 3 ...-- 4 ....-  5 .....
'       6 -.... 7 --... 8 ---.. 9 ----.  0 -----

'       FULL STOP          [.]  .-.-.-     COLON                   [:]  ---...
'       COMMA              [,]  --..--     SEMICOLON               [;]  -.-.-.
'       QUESTION MARK      [?]  ..--..     EQUAL SIGN (DOUBLE DASH)[=]  -...-
'       APOSTROPHE         [']  .----.     PLUS                    [+]  .-.-.
'       EXCLAMATION MARK   [!]  -.-.--     HYPHEN minus            [-]  -....-
'       FWD SLASH          [/]  -..-.      UNDERSCORE              [_]  ..--.-
'       PARENTHESIS open   [(]  -.--.      QUOTATION MARK          ["]  .-..-.
'       PARENTHESIS closed [)]  -.--.-     DOLLAR SIGN             [$]  ...-..-
'       AMPERSAND          [&]  .-...      AT SIGN (COMMAT) [WR]   [@]  .--.-.

'_______________________________________ENCODED MORSE CODE ALPHANUMERIC CHARACTER SET______________________________________
'       A   6,   B  17,   C  21,   D   9,   E   2,    F  20,
'       G  11,   H  16,   I   4,   J  30,   K  13,    L  18,
'       M   7,   N   5,   O  15,   P  22,   Q  27,    R  10,
'       S   8,   T   3,   U  12,   V  24,   W  14,    X  25,
'       Y  29,   Z  19, 

'       1  62,   2  60,   3  56,   4  48,   5  32, 
'       6  33,   7  35,   8  39,   9  47,   0  63

'       FULL STOP          	[.]  106     COLON                      [:]  71
'       COMMA              	[,]  115     SEMICOLON                  [;]  85
'       QUESTION MARK      	[?]  76      EQUAL SIGN (BREAK)	    [=]  49
'       APOSTROPHE         	[']  94      PLUS                       [+]  42
'       EXCLAMATION MARK   	[!]  117     HYPHEN (MINUS SIGN)        [-]  97
'       FORWARD SLASH      	[/]  41      UNDERSCORE                 [_]  108
'       PARENTHESIS (OPEN)   	[(]  45      QUOTATION MARK             ["]  82
'	  PARENTHESIS (CLOSED) 	[)]  109     DOLLAR SIGN          	    [$]  200
' 	  AMPERSAND          	[&]  34      AT (@) SIGN PROSIGN BAR WR [@]  86		SPACE [ ] 0

#no_data						'not uploading any EEPROM data

'__________________________________________________________________________________________________________________________

'user_setup
'__________________________________________________________________________________________________________________________

'**************************************************** NB **************************************************************
'for Morse code (CW) insert 2N3904 BJT into SG-8002 PTT Swiss female headers & remove BJT from FSK Swiss female headers
'for FSK remove PTT 2N3904 BJT & use wire link from SG-8002 GRN pin & circuit groundd (transmitter always on for FSKCW)
'**********************************************************************************************************************

'define CW/FSKCW beacon lookup table (callsign usually, tactcial for ISM bands or Aamteur Radio for ham bands)
'PICAXE BASIC macros create powerful user defined functions & allow you to pass, process & receive data back & forth

'you can combine beacon & grid to send in one transmission with fast CW & remove TX_GRID applicable code
#MACRO MY_BEACON(b6, b7)

	lookup b6,("ROM+"),b7		
	
#ENDMACRO

'if telemetry not being transmitted then transmit this Maidenhead grid square locator instead
#MACRO MY_GRID(b6, b7)

	lookup b6,("EN58+"),b7		
	
#ENDMACRO
 
'define only CW or FSKCW (only one mode at a time allowed)
'because of SG-8002's 5 minute warmup & frequency drift of 20 to 30 hertz (Hz) QRSS CW isn't used & fast CW is preferred
'fast CW more suitable on Amateur Radio bands in propagation beacon segments
#define CW						'on-off carrier keying (OOK or OOCK) CW
'#define FSKCW					'frequency shift keying (FSK} CW

'define only WPM or QRSS speed values (only one speed at a time allowed)
#define WPM 10					'words-per-minute (WPM) & speed for regular Morse code (CW) (change as required)
'#define QRSS 6					'QRSS mode dit time speed in seconds (s) (QRSS 6 is popular for weak signal work)

'comment out if telemetry not being added to CW/FSKCW beaon (have to write custom routines for this purpose)
#define TELEMETRY					'append telemery to end of CW/FSKCW beacon (air temp/battery voltage or whatever)

symbol V_COR = 10					'battery DC voltage (VDC) correction (cor) in ones & tenths (change as required) 
symbol T_COR = 30					'temperature cor in degs. Celsius (C) in ones & tenths(change as required)

'600 seconds is QRSS CW & FSKCW 10 min frame standard (for fast CW reduce to more appropriate beacon timing sequence)
symbol TX_FRAME = 600				'CW/FSKCW tranmsmission (tx) time frame in secs (change as required for mode)

symbol TM_FRAME = 3				'grid transmitted every TM_FRAME if telemetry sent (changed as required)	

'end user_setup
'__________________________________________________________________________________________________________________________

'program_functions

'convert ASCII numbers 0 to 9 to CW cut numbers (for faster tx) this table varies a bit depending on reference used	
'				0  1  2  3  4  5  6  7  8  9
'				T  A  U  V  H  E  B  G  D  N
#MACRO CUT_CW_NUMBER(b6, b7)
	
	b6 = b6 - 48				'convert ASCII value to decimal number
	
	lookup b6,("TAUVHEBGDN"),b7		'value returned using variable b7
	
#ENDMACRO

'transmit temperature & battery voltage values (as required)
'b3 to b5 are single digits to transmit, b6 is flag (0 or 1) if third digit transmitted (is not for temperatures)
#MACRO TX_TELEMETRY(b3, b4, b5, b6)

	CHARACTER = b3				'get first digit

	call TX_MORSE_CODE

	pause CHAR_SPAC
		
	CHARACTER = b4				'get second digit

	call TX_MORSE_CODE
	
	pause CHAR_SPAC
	
	b7 = b6					'get third digit transmit status flag
		
	if b7 = 1 then				'okay to transmit third digit?
		
		CHARACTER = b5			'yes
	
		call TX_MORSE_CODE
	
	endif

#ENDMACRO

'transmit Maidenhead grid square location if not transmitting telemetry
#MACRO TX_GRID

	pause WORD_SPAC
	
	INDEX = 0
		
	do
		
		MY_GRID(INDEX, CHARACTER)
		
		if CHARACTER = EOT then		'end-of-transmission character reached
			
			exit	
		
		endif
	
		if INDEX > 0 then
		
			pause CHAR_SPAC		'add character spacing after first but not after last character transmitted
	
		endif
		
		call TX_MORSE_CODE		'transmit encoded Morse character (non-EOT)
			
		inc INDEX
		
	loop

#ENDMACRO

'end program_functions
'__________________________________________________________________________________________________________________________

'program_setup

'first 4 words (8 bytes) w0, w1, w2, w3 (b0, b1, b2, b3, b4, b5, b6, b7) variables reserved for macros & maths
symbol J = b0					'general purpose loop variable

symbol INDEX = b8					'general purpose index & lookup tables index pointer

symbol CHARACTER = b9				'general purpose character (byte) variable

symbol FLAG = b10					'character bit even/odd status/flag variable

symbol ENCODED = b11				'encoded Morse code value variable

symbol GRID_FRAME = b12				'grid frame transmission counter

'ADC parameters used with voltage with added voltage divider circuit
symbol ADC_REF = b13				'ADC calibration reference

symbol TM_DRIFT = b14				'for adjusting time variable timing errors in ms

symbol TEMP_C = w11				'DS18B20 temperature 16-bit word

symbol VCC = w12					'VCC battery reference voltage word (16-bit) variable (b24/b25)

symbol VBAT = w13					'source battery voltage word (16-bit) variable (b26/b27)

symbol KEY = C.0					'key transmitter on/off (flash LED) with SG8002 2N3904 switch on ground pin	
symbol FSK = C.1					'FSKCW shift (space/mark) using 2N3906 switch connected to SG8002 OE/ST pin
symbol DS18B20 = C.2				'DS18B20 direct reading digital thermometer (degs. C)
symbol PUSH_BUTTON = pinC.3			'push button (if not connected program takes power on as "push button" press
symbol VBAT_PORT = C.4				'battery voltage level monitor control circuit connect

symbol EOT = "+"					'prosign bar AR ("+") end-of-transmission marker

symbol T_REF = 1200				'Morse code standard time reference in milliseconds (ms)

symbol TOTAL = 5					'total number of J loops (1 to 5) for averaging temperatures & voltages

#ifdef CW						'using "normal" speed (non-QRSS) Morse code continous wave (CW)
		
	symbol DIT_TIME = T_REF / WPM	'standard definition but depends on PICAXE clock error & may need to be adjusted

	symbol CHAR_SPAC = DIT_TIME * 3	'standard Morse code definition character spacing
	symbol WORD_SPAC = DIT_TIME * 6	'standard Morse code definition word spacing (7 is older standard) 
	
#endif

#ifdef QRSS
	
	symbol DIT_TIME = 1000 * QRSS		'QRSS dit time length in ms
	
	symbol CHAR_SPAC = DIT_TIME * 2	'shorten up character spacing for QRSS (still readable & faster tx time)
	symbol WORD_SPAC = DIT_TIME * 5	'shorten up word spacing for QRSS (still readable & faster tx time)
	
#endif

symbol DAH_TIME = DIT_TIME * 3		'standard Morse code definition for CW & FSKCW

#define KEY_DOWN high KEY			'define macros for CW key on/off carrier keying of SG8002 GND pin
#define KEY_UP low KEY

#define FSK_MARK high FSK			'define macros for FSK MARK & SPACE keying of SG8002 OE/ST pin
#define FSK_SPACE low FSK

'end program_setup
'__________________________________________________________________________________________________________________________

'initialize_program

' calibrate PICAXE ADC to actual VCC (never assume it's 5 VDC because it's usually not!)
calibadc ADC_REF					'calibrate ADC with internal reference voltage & determine actual VCC
	VCC = 26112 / ADC_REF * 100 / 256
	
#ifdef CW
	
	KEY_UP					'turn CW transmitter (carrier) off
	
#endif


#ifdef FSKCW
	
	KEY_UP					'transmitter LED indicator off
	FSK_SPACE					'shift SG8002 carrier down to space frequency (transmitter on continuously)
	
#endif

do

	'wait till user presses push button (ignored if push button isn't connected)
	if PUSH_BUTTON = 1 then			
		
		exit		

	endif

	'autostart if no press PB (ensures transmitter auto-restarts if PICAXE resets for some reason)
	if time >= TX_FRAME then
		
		exit
		
	endif
		
loop

'end initialize_program
'__________________________________________________________________________________________________________________________

'main_program

MAIN:
	
	disabletime					'stop PICAXE internal clock (required)
		
		time = 0				'reset PICAXE internal clock (every beacon transmit cycle)
			
	enabletime					'restart PICAXE internal clock (required)
	
	call TX_BEACON				'transmit beacon
	
	do						
	
	'clock will run slow or fast due to PICAXE internal parts-per-million (ppm) error
	'only way to remove timing error is with a DS3231 real-time-clock (RTC)
	
	loop until time = TX_FRAME		'wait til PICAXE interal time equals TX_FRAME
	
goto MAIN						'repeat tx sequence forever or until power removed

end

'end main_program
'__________________________________________________________________________________________________________________________

'program_subroutines

'CW subroutine transmits either Morse code CW, FSKCW beacon & any required telemetry 
TX_BEACON:
	
	INDEX = 0					'reset lookup table index pointer
	
	do
				
		'get next CW character from lookup table (CHARACTER is loaded with result)	
		MY_BEACON(INDEX, CHARACTER)		
		
		if CHARACTER = EOT then		'end-of-transmission character reached exit loop (AR not sent when using QRSS CW)
			
			exit	
		
		endif
	
		if INDEX > 0 then
		
			pause CHAR_SPAC		'add character spacing after first but not after last character transmitted
			
		endif
				
		call TX_MORSE_CODE		'transmit encoded Morse character (non-EOT)
			
		inc INDEX
	
	loop 						'keep going until EOT character triggers exit

	'drop here once EOT character reached & either transmit telemetry or grid square (only one or the other)
	'transmit grid square every 30 minutes regardless if transmitting telemetry
	#ifdef TELEMETRY
		
		if GRID_FRAME = TM_FRAME then	'tx grid square every TM_FRAME interval (skips a telemetry frame)
				
			TX_GRID
			
			GRID_FRAME = 0
			
		else
			
			call ADD_TELEMETRY						

			inc GRID_FRAME
			
		endif
	
	#else
		
		TX_GRID				'only sending grid sqare frame
			
	#endif

	#ifdef CW					'regular CW Morse mode
		
		CHARACTER = EOT			'for non-QRSS Morse code modes sent EOT character
			
		pause DAH_TIME
			
		call TX_MORSE_CODE
			
		KEY_UP				'turn carrier/transmitter off
			
	#endif					
	
return

'Routine to transmit Morse code characters using bit shifting 
'Based on orignal standard speed Morse code (CW) by Steve Paige, VK6HV                  
'Modified for extremely slow speed CW (QRSS) by Robert C. Mazur, VA3ROM (July 2016)     
TX_MORSE_CODE:  			
	
	'encode character for tx (converts from ASCII to specific Morse code encoded value)
	CHARACTER = CHARACTER - 32		'convert from ASCII
				
	'search for Morse code encoded value in lookup table
	lookup CHARACTER,(0,71,82,0,200,0,34,94,45,109,0,42,115,97,106,41,_
				63,62,60,56,48,32,33,35,39,47,71,85,0,49,0,76,86,_
				6,17,21,9,2,20,11,16,4,30,13,18,7,5,15,22,27,10,8,_
				3,12,24,14,25,29,19),ENCODED
	
	'check for word space character first (saves time)
	if ENCODED = 0 then
		
		pause WORD_SPAC

		return
	
	endif			

	'right bit shift (division) routine keeps going until Morse code character shifted value equals 1
	do

		FLAG = ENCODED % 2		'modulo function determines if value odd or even (1 or 0)
		
		#ifdef CW
				
			KEY_DOWN			'turn carrier/transmitter on
				
		#endif
		
		#ifdef FSKCW
			
			KEY_DOWN			'transmitter LED indicator on
			FSK_MARK			'shift carrier/transmitter frequency up (mark)
				
		#endif
				
		if FLAG = 1 then			'odd so do dah else even so do dit
					
			pause DAH_TIME
				
		else
		
			pause DIT_TIME
			
		endif
		
		#ifdef CW
		
			KEY_UP			'turn off carrier
	
		#endif
		
		#ifdef FSKCW
		
			KEY_UP			'transmitter LED indicator off
			FSK_SPACE			'shift carrier down (space)
		
		#endif
	
 		ENCODED = ENCODED / 2     	'division by 2 performs binary right bit shift
		
		if ENCODED <> 1 then		'add inter-character element spacing (if required)
			
			pause DIT_TIME
			
		endif
		
	loop until ENCODED = 1			'loop until all bits shifted & value 1 is reached

return						'back to main routine & get next character to send (if any)

'Optional subroutine measures external ambient air temperature & battery supply voltage
ADD_TELEMETRY:

	readtemp12 DS18B20, TEMP_C		'get current temperature from DS18B20 digital thermometer
	
	pause WORD_SPAC				'DS18B20 needs 500 to 750 ms to process data (convenient word spacing here)
	
	if TEMP_C > 255 then			'process negative temperatures
	  
		TEMP_C = TEMP_C / 2 or 128	'right bit shift & make positive (removes negative sign from value)
		TEMP_C = TEMP_C + 50		'offset result by 50 to indicate negative temperatures (-1C = 51, -49C = 99)
		
	else						'process positive temperature
	  
		TEMP_C = TEMP_C  / 2		
	
	endif
	
	TEMP_C = TEMP_C - T_COR			'correct error (+/-) using external thermometer (change as required)
	
	bintoascii TEMP_C, b0, b1, b2, b3, b4
	
	'convert ASCII numbers to Morse code "cut" CW characters (value returned by function in second variable)
	CUT_CW_NUMBER(b2, b5)
		CUT_CW_NUMBER(b3, b6)
			CUT_CW_NUMBER(b4, b7)
	
	b3 = b5					'shift out leading zero from temperature group (used for temps over 99.9C)
		b4 = b6
	
	TX_TELEMETRY(b3, b4, b5, 0)		'transmit temperature
	
	readadc VBAT_PORT, VBAT			'read analog voltage from resistive voltage divider
	
	'using ~3x voltage divider scaling factor (3:1 resistive voltage divider) +/- correction + 5/4 rounding
	VBAT = VBAT * VCC / 100 * 3 + V_COR + 5 / 10
	
	bintoascii VBAT, b0, b1, b2		
	
	CUT_CW_NUMBER(b0, b3)
		CUT_CW_NUMBER(b1, b4)
			CUT_CW_NUMBER(b2, b5)
	
	TX_TELEMETRY(b3, b4, b5, 1)		'transmit voltage (decimal point not transmitted - tenths of volt is 3rd digit)
	
return
'end program_subroutines
